"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from libs import base
from copy import deepcopy
from controllers import cluster_controller
from constants import MULTICLUSTER_UPDATE_TIMEOUT
from datetime import datetime
from itertools import chain
from constants import NOT_SUPPORTED


class MultiCluster(object):

    def __init__(
            self, cluster_obj, cluster_name, xdr_port):
        """ Initiate MultiCluster Object
        :param cluster_obj: Cluster Object
        :param cluster_name: Cluster Name
        :param xdr_port: XDR Port of AS Cluster
        :return: None
        """
        self._multicluster = {}  # List of all the clusters in a session
        # List also includes discovered clusters

        self._discovered_cluster_objects = {}  # Clusters whose cluster object
        # is present in session

        self._last_accessed = datetime.now()  # Required for update process
        self._xdr_port = xdr_port  # Common XDR port for all clusters

        self.add_cluster_object(cluster_obj, cluster_name, self._xdr_port)

    def remove_cluster_from_multicluster(self, cluster_id):
        """Remove the cluster from the list of session clusters"""
        if cluster_id in self._multicluster:
            cluster_obj = self._multicluster[cluster_id]
            if cluster_obj is not None:
                cluster_obj.close_sockets()

            del self._multicluster[cluster_id]

    def update_cluster_name(self, cluster_id, cluster_name):
        """Updates cluster name in the multicluster list"""
        if cluster_id in self._multicluster:
            self._multicluster[cluster_id]['cluster_name'] = cluster_name

    def get_multicluster_data(
            self, xdr_port, session, session_manager, cluster_manager):
        """ Return stats of all clusters in mc list
        :param xdr_port: XDR Port of all the monitored clusters
        :param session: {'session': <session-id>}
        :param session_manager: session manager object
        :param cluster_manager: cluster manager object
        :return: { <cluster-id/seednode>: <cluster-stats> }
        """
        self._last_accessed = datetime.now()
        """
        # Remove secured cluster if deleted from session (Logout)
        for cluster_id, cluster_obj in self._discovered_cluster_objects.items():
            response = session_manager.is_cluster_id_present(
                session['session'], cluster_id)
            if not response:
                # Remove only secured cluster so that user can login
                # with different credentials
                if cluster_obj.is_secured():
                    # Delete cluster from multicluster
                    del self._discovered_cluster_objects[cluster_id]
                    del self._multicluster[cluster_id]
                    # Refresh clusters containing the deleted cluster
                    for key, cluster_data in self._multicluster.items():
                        if cluster_id in cluster_data['xdr_info'].keys():
                            self.add_cluster_object(
                                key, self._discovered_cluster_objects[key],
                                cluster_data['cluster_name'], self._xdr_port
                            )
        """
        # If XDR port changes -> update entire mc
        if xdr_port != self._xdr_port:
            self._xdr_port = xdr_port
            self.update_multicluster(session, session_manager, cluster_manager)

        response = {}
        for cluster_id, cluster_data in self._multicluster.items():
            response[cluster_id] = dict()
            for key in cluster_data.keys():
                if key in ['dc_name', 'namespaces']:
                    # set() is not JSON serializable so converting to list
                    response[cluster_id][key] = list(cluster_data[key])
                elif key in ['xdr_port', 'cluster_key', 'cluster_size']:
                    pass  # Not required in response
                else:
                    response[cluster_id][key] = cluster_data[key]
        return response

    def get_discovered_cluster_objects(self):
        """ Return list of clusters in mc list who have cluster objects """
        return self._discovered_cluster_objects

    def add_cluster_object(self, cluster_obj, cluster_name, xdr_port):
        """ Add/Update cluster stats. Requires cluster object
        :param cluster_obj: Cluster Object
        :param cluster_name: Cluster Name
        :param xdr_port: XDR Port of Cluster
        :return: None
        """
        cluster_id = cluster_obj.get_cluster_id()
        user_credentials = cluster_obj.get_user()  # Get user credentials
        if user_credentials:
            user = user_credentials.values()[0]['username']
            password = user_credentials.values()[0]['password']
        else:
            user, password = None, None

        cluster_data = {  # Prepare dict
            cluster_id: dict(
                cluster_name=cluster_name,
                dc_name=set(),
                xdr_port=self._xdr_port,
                cluster_key=None,
                cluster_size=None,
                read_tps=dict(total=0, success=0),
                write_tps=dict(total=0, success=0),
                namespaces=set(),
                seednode=cluster_obj.get_seed(),
                nodes={},
                xdr_info={}
            )
        }
        cluster_info = cluster_data[cluster_id]
        node_obj_list = cluster_obj.get_node_id_node_obj()
        xdr_data = {}
        for node_id, node_obj in node_obj_list.items():
            status = node_obj.get_status()
            access_ip, access_port = node_obj.get_access_address().rsplit(
                ":", 1)
            cluster_info['nodes'][node_id] = {
                'ip': node_obj.get_ip(), 'port': node_obj.get_port(),
                'access_ip': access_ip, 'access_port': access_port,
                'status': status, 'lag': None,
                'cur_throughput': None, 'xdr_uptime': None
            }

            if not cluster_info['namespaces']:  # Obtain namespaces once
                cluster_info['namespaces'] = node_obj.get_namespace_list(
                    user, password)

            if status is "on":
                # Obtain cluster size once
                if cluster_info['cluster_size'] in [None, 'n/s']:
                    cluster_info['cluster_size'] = \
                        node_obj.get_statistic('cluster_size')
                # Obtain cluster key once
                if cluster_info['cluster_key'] in [None, 'n/s']:
                    cluster_info['cluster_key'] = \
                        node_obj.get_statistic('cluster_key')

                # Get read/write tps cluster wide
                read_tps = node_obj.get_read_tps()
                if read_tps['secondary']:
                    cluster_info['read_tps']['total'] += read_tps['secondary']
                    cluster_info['read_tps']['success'] += read_tps['y']

                write_tps = node_obj.get_write_tps()
                if write_tps['secondary']:
                    cluster_info['write_tps']['total'] += write_tps['secondary']
                    cluster_info['write_tps']['success'] += write_tps['y']

            # Get XDR data of every node
            xdr_obj = node_obj.get_xdr(xdr_port, user=user, password=password)
            if xdr_obj:
                xdr_stat = xdr_obj.get_subset_of_statistics(
                    ['timediff_lastship_cur_secs', 'cur_throughput', 'xdr_timelag',
                     'xdr-uptime', 'xdr_uptime'],
                    user=user, password=password)[0]

                if xdr_stat['timediff_lastship_cur_secs'] in [None, NOT_SUPPORTED]:
                    cluster_info['nodes'][node_id]['lag'] = xdr_stat['xdr_timelag']
                else:
                    cluster_info['nodes'][node_id]['lag'] = xdr_stat['timediff_lastship_cur_secs']

                cluster_info['nodes'][node_id]['cur_throughput'] = xdr_stat['cur_throughput']
                # Get xdr uptime
                if xdr_stat['xdr_uptime'] in [NOT_SUPPORTED, None]:
                    xdr_uptime = xdr_stat['xdr-uptime']
                else:
                    xdr_uptime = xdr_stat['xdr_uptime']

                cluster_info['nodes'][node_id]['xdr_uptime'] = xdr_uptime

                dc_config = xdr_obj.get_dc_config(user=user, password=password)

                if not dc_config:
                    continue

                xdr_data.update(dc_config)

        # Add XDR entry to cluster data shipping list
        for dc_config in xdr_data.values():
            node_address = ''.join([dc_config['seednode']['ip'], ':',
                                    dc_config['seednode']['port']])

            if node_address in cluster_info['xdr_info']:
                cluster_info['xdr_info'][node_address][
                    'shipping_namespaces'] += dc_config['namespaces']
            else:
                cluster_info['xdr_info'][node_address] = {
                    'shipping_namespaces': dc_config['namespaces']
                }

        self.add_xdr_cluster(xdr_data)  # Add xdr entry to multicluster list

        cluster_info['discovery'] = "complete"
        self.add_to_multicluster(cluster_data)
        self._discovered_cluster_objects[cluster_id] = cluster_obj

    def add_to_multicluster(self, cluster_data):
        """ Validate & map data before inserting into multicluster list
        :param cluster_data: Expected Format ->
        {
            '<cluster_key/ip-address>': {
                'cluster_name': <cluster-name>,
                'dc_name': {<dc-name>, },  # set()
                'namespaces': [<namespace>, ],  #set()
                'seednode': <seednode>,
                'nodes': {
                    <node-id/node-ip>: {
                        'ip': <ip>,
                        'port': <port>,
                        'status': <on/off>,
                        'lag': <lag>,
                        'cur_throughput':<cur_throughput>,
                        'access_ip': <access_ip>,
                        'access_port': <access_port>
                    }
                },
                'read_tps': {
                    'total': <total_read_tps>, 'success': <success_read_tps>
                },
                'write_tps': {
                    'total: <total_write_tps>, 'success': <success_write_tps>
                }
                'xdr_info': {
                    <cluster_key/ip-address>:{
                        'shipping_namespaces': [<namespace>]
                    }
                }
                'discovery': <in_progress/complete/secured/not_available>
            }
        }
        :return: None
        """

        cluster_id = cluster_data.keys()[0]
        cluster_info = cluster_data.values()[0]

        for node_info in cluster_info['nodes'].values():
            node_address = ''.join(
                [node_info['ip'], ':', node_info['port']])

            external_node_address = ''.join(
                [node_info['access_ip'], ':', node_info['port']])

            # Check if this node_address is present in other cluster's
            # xdr_info, If yes then replace node_address with cluster_key
            for mc_id, mc_info in self._multicluster.items():
                for shipping_ip in mc_info['xdr_info'].keys():
                    if shipping_ip in [node_address, external_node_address]:
                        if cluster_id != shipping_ip:
                            # This is when cluster id is uuid and shipping
                            # address in other clusters node is IP address
                            # Get shipping namespaces
                            if cluster_id in mc_info['xdr_info'].keys():
                                mc_info['xdr_info'][cluster_id][
                                    'shipping_namespaces'] = list(
                                    set(
                                        mc_info['xdr_info'][
                                            cluster_id]['shipping_namespaces']
                                    ).union(
                                        mc_info['xdr_info'][
                                            shipping_ip]['shipping_namespaces']
                                    )
                                )
                                mc_info['xdr_info'].pop(shipping_ip)
                            else:
                                mc_info['xdr_info'][cluster_id] = mc_info[
                                    'xdr_info'].pop(shipping_ip)

                # Check if this node_address is present in multicluster
                # set, absorb dc names and then delete that entry
                # if node_address == mc_id:
                if mc_id in [node_address, external_node_address]:
                    cluster_info['dc_name'].update(mc_info['dc_name'])
                    del self._multicluster[mc_id]

        if cluster_info['xdr_info']:
            for address, info in cluster_info['xdr_info'].items():
                marked = False

                # Check if this address is present in discovered
                # cluster's node list, If yes then replace
                # this entry with the cluster key of that node address
                if address in self._multicluster.keys():
                    continue
                changed = False
                for mc_id, mc_info in self._multicluster.items():
                    if marked:
                        break

                    cluster_obj = self._discovered_cluster_objects.get(
                        mc_id, None)
                    for node_info in mc_info['nodes'].values():

                        if marked:
                            break

                        address_list = set()
                        node = ''.join(
                            [node_info['ip'], ":", node_info['port']])
                        address_list.add(node)
                        address_list.add(''.join(
                            [node_info['access_ip'], ":",
                             node_info['access_port']]))

                        if cluster_obj:
                            node_obj = cluster_obj.get_node(node)
                            if node_obj:
                                address_list.update(
                                    node_obj.get_alternate_address())

                        if address in address_list:
                            if mc_id != address:
                                cluster_info['xdr_info'][mc_id] = cluster_info[
                                    'xdr_info'].pop(address)
                                changed = True
                                marked = True
                            else:
                                marked = True

                if not changed:
                    # Check if the cluster is shipping to itself
                    # This is a possible scenario
                    for nodes in cluster_info['nodes'].values():
                        node_address = ''.join(
                            [nodes['ip'], ':', nodes['port']])
                        external_node_address = ''.join(
                            [nodes['access_ip'], ':', nodes['port']])
                        if address in [node_address, external_node_address]:
                            if cluster_id != address:
                                cluster_info[
                                    'xdr_info'][cluster_id] = cluster_info[
                                        'xdr_info'].pop(address)

        self._multicluster.update(cluster_data)

    def add_xdr_cluster(self, xdr_data):
        """ This method checks if the shipping dc is already present
         and accordingly marks this link. If not, a new entry
        in multicluster list is created.

        :param xdr_data: This accepts data received from get_dc_config() of
        libs/base.py file.
        :return: None
        """

        for key, value in xdr_data.items():
            already_present = False

            found = None
            for nodes in value['Nodes']:
                node_address = ''.join([nodes['ip'], ':', nodes['port']])
                # If entry for same node already exists
                if node_address in self._multicluster.keys():
                    self._multicluster[node_address]['dc_name'].add(key)
                    already_present = True

                if already_present:
                    break

                # Check if cluster object containing this node exists
                for mc_id, mc_info in self._multicluster.items():
                    for address in mc_info['nodes'].values():
                        node = ''.join(
                            [address['ip'], ':', address['port']])

                        # Absorb dc_name
                        if node_address == node:
                            mc_info['dc_name'].update(mc_info['dc_name'])
                            already_present = True
                            found = mc_id
                            break

                if already_present:
                    break  # Even if one node is found stop further search

            if already_present:
                # Check if there are any new nodes that are not present in
                # this existing cluster's list. If yes, then absorb these nodes
                if not found:
                    continue
                cluster_info = self._multicluster.get(found, None)
                if not cluster_info:
                    continue
                cluster_obj = self._discovered_cluster_objects.get(found, None)

                node_address_list = []
                for nodes in cluster_info['nodes'].values():
                    node_address_list.append(''.join([
                        nodes['ip'], ':', nodes['port']]))

                for nodes in value['Nodes']:
                    node_address = ''.join([nodes['ip'], ':', nodes['port']])
                    if node_address not in node_address_list:
                        access_ip = nodes['ext_ip'] if \
                            nodes['ext_ip'] else nodes['ip']
                        access_address = ''.join(
                            [access_ip, ':', nodes['port']])

                        if cluster_obj:
                            # Get user credentials
                            user_credentials = cluster_obj.get_user()
                            if user_credentials:
                                user = user_credentials.values()[0][
                                    'username']
                                password = user_credentials.values()[0][
                                    'password']
                            else:
                                user, password = None, None

                            cluster_obj.add_existing_node(
                                access_address, user, password)
                        else:
                            cluster_info['nodes'].update({
                                node_address: {
                                    'ip': nodes['ip'],
                                    'port': nodes['port'],
                                    'status': 'on',  # Need to validate this
                                    'cur_throughput': None,
                                    'access_ip': nodes['ext_ip'],
                                    'access_port': nodes['port']
                                }
                            })

                # Take next dc for processing as this dc is
                # already present in mc list
                continue

            # If code reaches here that means this dc is new and an entry
            # can be created in mc list for this dc

            # Use External IP if available for info calls
            access_ip = value['seednode']['ext_ip'] if \
                value['seednode']['ext_ip'] else value['seednode']['ip']
            seednode = ''.join([access_ip, ':',
                                value['seednode']['port']])
            node_available = base.get_node_availabilty(
                access_ip, value['seednode']['port'])

            if node_available == 'available':
                node_available = 'in_progress'

            cluster_data = {
                seednode: dict(
                    cluster_name=None,
                    # Can be replaced with set literal but set literal is
                    # not supported in python 2.6
                    dc_name=set([key, ]),
                    xdr_port=self._xdr_port,
                    cluster_key=None,
                    cluster_size=None,
                    read_tps=dict(total=0, success=0),
                    write_tps=dict(total=0, success=0),
                    namespaces=set(),
                    seednode=seednode,
                    nodes={},
                    xdr_info={},
                    discovery=node_available
                )
            }

            cluster_info = cluster_data.values()[0]
            # Add remaining nodes to cluster
            for node_info in value['Nodes']:
                node_address = ''.join([node_info['ip'], ':',
                                        node_info['port']])
                if node_address not in cluster_info['nodes']:
                    access_ip = node_info['ext_ip'] if \
                        node_info['ext_ip'] else node_info['ip']
                    node_available = base.get_node_availabilty(
                        access_ip, node_info['port'])
                    status = 'on' if node_available in [
                        'available', 'secured'] else 'off'

                    cluster_info['nodes'].update({
                        node_address: {
                            'ip': node_info['ip'], 'port': node_info['port'],
                            'access_ip': access_ip,
                            'access_port': node_info['port'],
                            'status': status,
                            'lag': None, 'cur_throughput': None,
                            'xdr_uptime': None
                        }
                    })

            self.add_to_multicluster(cluster_data)

    def update_multicluster(self, session, session_manager, cluster_manager):
        """
        This method will update multicluster stats. This method will also create
        new cluster objects if a new dc is available.

        :param session: session of user. Expected Format ->
        {'session': <session-id>}
        :param session_manager: session manager object which is requried by
        get_cluster_id() of /controllers/cluster_controller.py
        :param cluster_manager: cluster manager object which is required by
        get_cluster_id() of /controllers/cluster_controller.py
        :return: None
        """

        # Stop update process if user has not requested for multicluster view
        # for a certain period
        time_diff = datetime.now() - self._last_accessed
        if time_diff.seconds > MULTICLUSTER_UPDATE_TIMEOUT*60:
            return

        # Create cluster objects of entries in mc list that are available
        for cluster_id, cluster_info in self._multicluster.items():
            if cluster_info['discovery'] is 'in_progress':

                node_list = []
                for node_address, nodes in cluster_info['nodes'].items():
                    node_list.append({
                        'ip': nodes['ip'], 'port': nodes['port'],
                        'access_ip': nodes['access_ip'],
                        'access_port': nodes['access_port']
                    })

                for node_info in node_list:
                    if node_info['access_ip']:
                        ip, port = node_info['access_ip'], \
                                   node_info['access_port']
                    else:
                        ip, port = node_info['ip'], node_info['port']

                    node_available = base.get_node_availabilty(ip, port)
                    if node_available == 'available':
                        # Delete current entry as this datacenter will now be
                        # represented by a new cluster object
                        del self._multicluster[cluster_id]

                        seed_node = ''.join([ip, ':', port])
                        cluster_details = {
                            'seed_node': seed_node,
                            'node_list': node_list
                        }

                        # Create Cluster Object
                        response = cluster_controller.get_cluster_id(
                            cluster_details, session, session_manager,
                            cluster_manager
                        )

                        cluster_id = response.get('cluster_id', None)
                        if cluster_id:
                            cluster_obj = cluster_manager.get_cluster_by_id(
                                cluster_id)
                            # Add this cluster to mc list
                            self.add_cluster_object(
                                cluster_obj, None, self._xdr_port)
                            break

        # Using deepcopy to avoid RuntimeError while changing size of
        # multicluster dict while iterating over it
        # Running update for multiclusters in sequence of clusters with objects
        # and then with clusters with no objects. Here, objects mean instance
        # of Class Cluster
        object_clusters = {}
        non_object_clusters = {}

        for key in self._multicluster.keys():
            if key in self._discovered_cluster_objects.keys():
                object_clusters[key] = deepcopy(self._multicluster[key])
            else:
                non_object_clusters[key] = deepcopy(self._multicluster[key])

        for cluster_id in chain(
                object_clusters.keys(), non_object_clusters.keys()):
            cluster_obj = self._discovered_cluster_objects.get(
                cluster_id, None)

            try:
                cluster_info = self._multicluster[cluster_id]
            except KeyError:
                continue

            if cluster_obj:
                self.checks_for_cluster_with_objects(cluster_obj, cluster_info)
            else:
                node_info = cluster_info['nodes'].values()[0]
                access_ip = node_info['access_ip'] if \
                    node_info['access_ip'] else node_info['ip']
                port = node_info['port']
                node_available = base.get_node_availabilty(access_ip, port)

                # Set to in_progress to create object in next update call
                if node_available is 'available':
                    cluster_info['discovery'] = "in_progress"

        del object_clusters, non_object_clusters

    def checks_for_cluster_with_objects(self, cluster_obj, cluster_info):
        """ Checks for cluster who have objects
        :param cluster_obj: Cluster object
        :param cluster_info: Cluster stats
        :return: None
        """
        kwargs = {}
        user_credentials = cluster_obj.get_user()  # Get user credentials
        if user_credentials:
            kwargs['user'] = user_credentials.values()[0]['username']
            kwargs['password'] = user_credentials.values()[0]['password']
        else:
            kwargs['user'], kwargs['password'] = None, None

        cluster_obj.get_updates(**kwargs)  # Update cluster

        new_cluster_key = None
        new_cluster_size = None
        state_changed = False
        if self._xdr_port != cluster_info['xdr_port']:
            state_changed = True

        # Calculate read/write tps
        cluster_info['read_tps'].update({'total': 0, 'success': 0})
        cluster_info['write_tps'].update({'total': 0, 'success': 0})

        for node_id, node_obj in cluster_obj.get_node_id_node_obj().items():
            if state_changed:  # For xdr port change
                break
            if node_id not in cluster_info['nodes'].keys():
                state_changed = True
                break

            status = node_obj.get_status()
            if cluster_info['nodes'][node_id]['status'] != status:
                state_changed = True
                break

            if status == 'on':
                read_tps = node_obj.get_read_tps()
                if read_tps['secondary']:
                    cluster_info['read_tps']['total'] += read_tps['secondary']
                    cluster_info['read_tps']['success'] += read_tps['y']

                write_tps = node_obj.get_write_tps()
                if write_tps['secondary']:
                    cluster_info['write_tps']['total'] += write_tps['secondary']
                    cluster_info['write_tps']['success'] += write_tps['y']

                if new_cluster_size in [None, 'n/s']:  # Check once
                    new_cluster_size = node_obj.get_statistic('cluster_size')
                    if cluster_info['cluster_size'] in [None, 'n/s']:
                        cluster_info['cluster_size'] = new_cluster_size
                    elif cluster_info['cluster_size'] != new_cluster_size:
                        state_changed = True
                        break

                if new_cluster_key in [None, 'n/s']:
                    new_cluster_key = node_obj.get_statistic('cluster_key')
                    if cluster_info['cluster_key'] in [None, 'n/s']:
                        cluster_info['cluster_key'] = new_cluster_key
                    elif cluster_info['cluster_key'] != new_cluster_key:
                        state_changed = True
                        break

                if not cluster_info['namespaces']:  # Get namespaces if null
                    cluster_info['namespaces'] = node_obj.get_namespace_list(
                        **kwargs)

            xdr_obj = node_obj.get_xdr(cluster_info['xdr_port'], **kwargs)
            if xdr_obj:
                xdr_stat = xdr_obj.get_subset_of_statistics(
                    ['timediff_lastship_cur_secs', 'cur_throughput', 'xdr_timelag',
                     'xdr-uptime', 'xdr_uptime'], **kwargs)[0]

                # Get xdr uptime
                xdr_uptime = xdr_stat['xdr-uptime'] if \
                    xdr_stat['xdr_uptime'] in [NOT_SUPPORTED, None] else \
                    xdr_stat['xdr_uptime']

                xdr_availability = False if xdr_uptime in [None, 'n/s'] else \
                    True

                old_xdr_uptime = cluster_info['nodes'][node_id]['xdr_uptime']

                # Earlier XDR was up and now it is down
                if old_xdr_uptime not in [None, 'n/s'] and \
                        not xdr_availability:
                    state_changed = True
                    break
                # Earlier XDR was down and now it is up
                elif old_xdr_uptime in [None, 'n/s'] and xdr_availability:
                    state_changed = True
                    break

                if xdr_uptime not in [None, 'n/s']:
                    try:
                        if int(xdr_uptime) < 5:  # Check if XDR was restarted
                            state_changed = True
                            break
                    except ValueError:
                        pass
                cluster_info['nodes'][node_id][
                    'xdr_uptime'] = xdr_uptime

                # If previous values of lag and cur_throughput are not set then
                # set them
                lag = xdr_stat['timediff_lastship_cur_secs']
                if lag in [None, NOT_SUPPORTED]:
                    lag = xdr_stat['xdr_timelag']

                if cluster_info['nodes'][node_id]['lag'] in [None, 'n/s'] or cluster_info['nodes'][node_id]['cur_throughput'] in [None, 'n/s']:
                        cluster_info['nodes'][node_id]['lag'] = lag
                        cluster_info['nodes'][node_id]['cur_throughput'] = xdr_stat['cur_throughput']
                # If previous values are set and do not match with the current
                # values then cluster state is changed
                elif cluster_info['nodes'][node_id]['lag'] != lag or cluster_info['nodes'][node_id]['cur_throughput'] != xdr_stat['cur_throughput']:
                        state_changed = True
                        break

        if state_changed:  # Update this cluster
            self.add_cluster_object(
                cluster_obj, cluster_info['cluster_name'], self._xdr_port)

    def get_cluster_ipmap(self, seednode):
        """
        Return nodes of cluster with int-ext-ipmap if present
        This is used by get_cluster_id() of /controllers/cluster_controller.py
        :param seednode: Seed address of cluster <IP:PORT>
        :return: [{
                    'ip': <ip>,
                    'port': <port>,
                    'access_ip': <access-ip>,
                    'access_port': <access-port>
                }, ]
        """
        node_list = []
        if not seednode:
            return node_list

        found = None
        for cluster_id, cluster_info in self._multicluster.items():
            if found:
                break
            for node in cluster_info['nodes'].values():
                node_address = ''.join([node['ip'], ':', node['port']])
                alternate_node_address = \
                    ''.join([node['access_ip'], ':', node['access_port']]) if \
                    node['access_ip'] else None

                if seednode in [node_address, alternate_node_address]:
                    found = cluster_id
                    break

        if not found:
            return node_list

        cluster_info = self._multicluster.get(found, None)

        for node in cluster_info['nodes'].values():
            node_list.append({
                'ip': node['ip'], 'port': node['port'],
                'access_ip': node['access_ip'],
                'access_port': node['access_port']
            })

        return node_list
